# -*- coding: utf-8 -*-
"""
Graphs: traversals - BFS and DFS
"""

from algopy import graph, graphmat, queue

#------------------------------------------------------------------------------
# ex 2.1

"""
Breadth-first search (BFS)
"""

# simple BFS, vertices are marked with booleans, with adjacency matrix
# display vertices, on line per tree

def __BFS(Gmat, s, M):
    q = queue.Queue()
    q.enqueue(s)
    M[s] = True
    while not q.isempty():
        s = q.dequeue()
        print(s)
        for adj in range(Gmat.order):
            if Gmat.adj[s][adj]:    # adj is adjacent to s
                if not M[adj]:  # adj is not marked
                    M[adj] = True
                    q.enqueue(adj)
                

def BFS(Gmat):
    M = [False] * Gmat.order
    for s in range(Gmat.order):
        if not M[s]:    #s is not marked
            __BFS(Gmat, s, M)
        print()


# BFS the builds the parent vector, used to mark vertices, 
# with adjacency lists representation

def __BFS_forest(G, s, p):
    q = queue.Queue()
    q.enqueue(s)
    p[s] = -1
    while not q.isempty():
        s = q.dequeue()
        for adj in G.adjlists[s]:
            if p[adj] == None:
                q.enqueue(adj)
                p[adj] = s

def BFS_forest(G):
    p = [None] * G.order
    for s in range(G.order):
        if p[s] is None:
            __BFS(G, s, p)
    return p    # represents the spanning forest
    
    
#------------------------------------------------------------------------------
# ex 2.2 q2.4

"""
Depth first traversal (DFS)
"""

# simple DFS with adjacency lists
def __dfs(G, s, M):
    M[s] = True    # usually vertices are marked here
    print(s, end=' ')
    for adj in G.adjLists[s]:
        if not M[adj]: 
            __dfs(G, adj, M)
                
def dfs(G):
    M = [False] * G.order
    for s in range(G.order):
        if not M[s]:
            __dfs(G, s, M)
        print()

            
# q.3(c) graph depth-first traversal with back edge detection (with adjacency matric)
# vertices are marked with parents
            
def __dfsForest(Gmat, s, p):
    for adj in range(Gmat.order):
        if Gmat.adj[s][adj]:
            if p[adj] == None: 
                p[adj] = s      # vertices have to be marked here
                print(s, "->", adj)     # tree edge
                __dfsForest(Gmat, adj, p)
            else:
                if adj != p[s]:
                    print(s, '->', adj, "back edge")    
                    # unless adj -> s is a back edge! *
                
def dfsForest(Gmat):
    p = [None] * Gmat.order
    for s in range(Gmat.order):
        if p[s] == None:
            p[s] = -1
            __dfsForest(Gmat, s, p)
    return p

# *vertices can be marked with their depths in trees: allows to detect back ede only once!


# q.4(c) digraph depth-first traversal -> prefix and suffix numbering with a single counter
# to detect edge types (with adjacency list representation)

def __dfs_digraph(G, s, pref, suff, cpt):
    cpt += 1
    pref[s] = cpt
    for adj in G.adjLists[s]:
        if pref[adj] == 0:
           # s -> adj: tree edge
           cpt = __dfs_digraph(G, adj, pref, suff, cpt)
        else:
            if pref[s] < pref[adj]:
                print (s, "->",  adj, "forward")
            else:
                if suff[adj] == 0:
                    print (s, "->",  adj, "back")
                else:
                    print (s, "->",  adj, "cross")
    cpt += 1
    suff[s] = cpt
    return cpt

# reminder: int values are "unmutable" => cpt cannot be passed by reference
# that why the functions returns cpt!

def dfs_digraph(G):
    pref = [0] * G.order
    suff = [0] * G.order 
    cpt = 0
    for s in range(G.order):
        if pref[s] == 0:
            cpt = __dfs_digraph(G, s, pref, suff, cpt)
    return(pref, suff)
